#! /opt/perl/bin/perl
#
# $Id: griphyn.pl,v 1.7 2001/10/02 20:37:18 voeckler Exp voeckler $
#
require 5.6.0;
use strict;
use DBI;			# database independent module
use IO::File;

# prepare to use scriptdir local modules
use File::Basename;		# import dirname() function
use lib dirname($0);		# add script dir to module search path
use Wrapper;			# local wrapper around dbase class
use Dagger;

# check of DBD::Pg
die "unable to access PostGreSQL driver!\n"
    if ( (grep 'Pg', DBI->available_drivers) == 0 );

# create dbase access instance, will die on error
my $dbh = Wrapper->new( 'voeckler', 'voeckler', 'voeckler' );
my %functioncall = 
    ( begin => [ \&myBegin, "<xid> | <exec>" ],
      file => [ \&myFile, "[i|o] <lfn>" ],
      arg => [ \&myArg, "<str>" ],
      stdin => [ \&myStdIn, "<lfn>" ],
      stdout => [ \&myStdOut, "<lfn>" ],
      stderr => [ \&myStdErr, "<lfn>" ],
      end => [ \&myEnd, "" ],
      cancel => [ \&myCancel, "" ],
      rc => [ \&myRC, "<lfn> <pfn>" ],
      get => [ \&myGet, "<lfn> -- will become obsolete!" ],
      dag => [ \&myDag, "<dagpfnbase> <lfn>" ],
      debug => [ \&myDebug, "<type> <level>" ],
      help => [ \&myHelp, "" ],
      set => [ \&mySet, "<key> <value>" ],
      unset => [ \&myUnset, "<key>" ],
      show => [ \&myShow, " | <key> ..." ],
      load => [ \&myLoad, "<pfn>" ],
      save => [ \&mySave, "<pfn>" ],
      dump => [ \&myDump, "<table>" ]
    );

sub processFile($$);
do { processFile( $dbh, shift(@ARGV) ) } while ( @ARGV );
warn "-- info: uncommitted modifications pending!\n" if $dbh->count;
exit 0;

#
# --------------------------------------------------------------------
#
sub processFile ($$) {
    # purpose: main loop
    # paramtr: $fn (IN): filename of file to read, or "<&STDIN"
    # returns: number of non-trivial lines processed.
    my $dbh = shift;
    my $fn = shift || '-';
    my $nr = 0;
    my $fh = new IO::File '<' . $fn;
    if ( defined $fh ) {
	while ( <$fh> ) {
	    s/\#.*//;		# remove comments
	    s/^\s+//;		# remove initial spaces
	    s/\s*\r?\n$//;	# remove trailing whitespaces
	    next unless length;	# skip empty lines
	    $nr++;		# count important lines

	    # obtain command separate from its argument
	    my ($cmd,$arg) = split /\s+/,$_,2;
	    $cmd = lc($cmd);
	    if ( exists $functioncall{$cmd} ) {	
		my $rv = $functioncall{$cmd}->[0]->($dbh,$arg);
		print "$rv\n" if defined $rv;
	    } else {
		last if $cmd eq 'exit';
		print "-- error: unknown command $cmd\n";
	    }
	}
	$fh->close();
    } else {
	print "-- error: open $fn: $!\n";
    }
    $nr;
}

sub myForce {
    # purpose: force to add something to the database
    # paramtr: $arg (IN): parameter containing cmd word
    # returns: cmd return value, or undef in case of error
    my $dbh = shift;
    my ($cmd,$arg,$rv) = split(/\s+/,shift,2);
    $cmd = lc($cmd);

    if ( exists $functioncall{$cmd} ) {
	$dbh->increment_force;
	$rv = $functioncall{$cmd}->[0]->($dbh,$arg);
	$dbh->decrement_force;
    } else {
	exit(0) if $cmd eq 'exit';
	print "-- error: unknown command $cmd!\n";
    }
    $rv;
}

sub myBegin ($) {
    # purpose: add a function specification to TRANSFORMATION
    # paramtr: $arg (IN): value of parameter
    # returns: value of XID to use, undef in case of error
    # !fixme!: no environment (yet)
    my $dbh = shift;
    my ($exec,$args) = split( /\s+/, shift, 2 );
    my $xid;

    # make sure that we start a clean transaction
    if ( defined $dbh->{m_transformation} ) {
	print "-- error: transaction already started\n";
	return undef;
    }

    if ( $exec =~ /^(\d+)$/ ) {
	# re-incarnate an existing function with different parameters
	print "-- info: recycling $xid\n";
	$xid = $1;
    } else {
	# maybe we alread know about this function
	my $r = $dbh->fetch( q{select XID from TRANSFORMATION 
			       where EXECUTABLE = ?}, $exec );
	if ( ref $r &&  @$r > 0 && ! $dbh->force ) {
	    # function already known, recycle id
	    print "-- warn: multiple matching entries, taking first\n" 
		if ( @$r > 1 );
	    $xid = $r->[0]->[0];
	    print "-- info: function already in database as fid=$xid\n";
	} else {
	    # data unknown, create new entry
	    # obtain next sequence number to use as key
	    $xid = $dbh->get_sequence('SEQ_XFORM') or
		return undef; # FIXME: die?
	    print "-- info: using fid=$xid\n";

	    # create atomic transaction to insert filename
	    return undef 
		unless $dbh->modify( q{insert into TRANSFORMATION(XID,EXECUTABLE)
				       values (?,?)}, $xid, $exec );
	    print "-- info: inserted $xid => $exec $args\n";
	}
    }

    # obtain new derived data id (ddid) to use for this transaction
    my $ddid = $dbh->get_sequence('SEQ_DDATA') or
	return undef;		# FIXME: die?

    # save xid in transaction, reset argument counter, store ddid
    $dbh->{m_transformation} = { xid => $xid, count => 0, ddid => $ddid };

    # return parameter id to be used for dependencies
    $xid;
}

sub myEnd {
    # purpose: commit open transaction
    # returns: number of committed insertions, undef for errors
    my $dbh = shift;
    unless ( defined $dbh->{m_transformation} ) {
	print "-- error: no block to finish\n";
	return undef;
    }

    my $result;
    if ( $result = $dbh->commit ) {
	print( "-- info: committed $result modifications.\n" );
    } else {
	print( "-- error: unable to commit $result modifications: ",
	       $dbh->handle->errstr, "\n" );
    }
    undef $dbh->{m_transformation};
    $result;
}

sub myCancel {
    # purpose: undo open transaction
    # returns: number of undone insertions, undef for errors
    my $dbh = shift;
    unless ( defined $dbh->{m_transformation} ) {
	print "-- error: no block to finish\n";
	return undef;
    }

    my $result;
    if ( $result = $dbh->rollback ) {
	print( "-- info: dropped $result modifications.\n" );
    } else {
	print( "-- error: unable to roll back $result modifications: ",
	       $dbh->handle->errstr, "\n" );
    }
    undef $dbh->{m_transformation};
    $result;
}

sub myHelp {
    # purpose: show available commands
    foreach my $k ( sort keys %functioncall ) {
	printf( "-- %-10s %s\n", $k, $functioncall{$k}->[1] );
    }
    undef;
}

sub myDump {
    # purpose: show content of tables
    my $dbh = shift;
    my @tables = map { lc($_) } $dbh->handle->tables;
    my %xlate = ( '\*' => 'TRANSFORMATION PARAMETER REPLICA_CATALOG DERIVED',
		  'file' => 'PARAMETER',
		  'rc' => 'REPLICA_CATALOG' );

    # undo shortcuts
    my $s = shift;
    foreach my $k ( keys %xlate ) {
	$s =~ s/$k/$xlate{$k}/g;
    }
    print "-- debug: $s\n";

    # show all tables
    my $count = 0;
    foreach my $arg ( map { lc($_) } split(/\s+/,$s) ) {
	# check, if table exists, before trying to access it
	if ( grep( /^$arg$/, @tables ) == 0 ) {
	    print "-- warn: $arg is not a known table name\n";
	    next;
	}

	# fetch all from named table
	my $r = $dbh->fetch( qq{select * from $arg} );
	if ( defined $r && ref $r ) {
	    print uc($arg), "\n";
	    foreach my $line ( @$r ) {
		print ' |', DBI::neat_list($line,30,'|'), "|\n";
		++$count;
	    }
	    print "\n";
	}
    }
    $count;
}

sub myDebug {
    # purpose: set level for debug messages
    # paramtr: usually kv-pairs
    my $dbh = shift;
    my @arg = split(/\s+/,shift);
    
    my ($k,$v);
    while ( @arg ) {
	$k = shift(@arg);
	$v = shift(@arg);
	if ( $v ) { $dbh->debug->{$k} = $v }
	else { delete $dbh->debug->{$k} }
    }

    while ( ($k,$v) = each %{$dbh->debug} ) {
	print "-- debug: $k=$v\n";
    }

    undef;
}

sub internalArg {
    # purpose: add any kind of argument PARAMETER and its dependencies, too.
    # paramtr:
    # returns: value of PID, undef in case of error
    my ( $dbh, $str, $iox, $pid ) = @_;

    # get position and update parameter count
    # mind, stdio does not use a valid position, set to -1 as indication
    my $pos = -1;
    $pos = $dbh->{m_transformation}->{count}++
	unless ( $iox eq 'I' or  $iox eq 'O' or  $iox eq 'E' );

    # phase 1: add non-file argument to database
    # maybe the argument is already in the database?
    my $r = $dbh->fetch( q{select PID from PARAMETER where value = ?}, 
			 $str );
    if ( ref $r && @$r > 0 && ! $dbh->force ) {
	# data already known, recycle argument
	print "-- warn: multiple matching entries, taking first\n" 
	    if ( @$r > 1 );
	$pid = $r->[0]->[0];
	print "-- info: data already in database at pid=$pid\n";
    } else {
	# data unknown, create new entry
	# obtain next sequence number to use as key
	$pid = $dbh->get_sequence('SEQ_PARAM') || return undef; # FIXME: die?
	print "-- info: using pid=$pid\n";

	# create atomic transaction to insert filename
	return undef 
	    unless $dbh->modify( q{insert into PARAMETER(PID,VALUE) 
				   values (?,?)},
				 $pid, $str );
	print "-- info: inserted $pid => $str\n";
    }

    # phase 2: add dependency between xid and pid to database
    # we are specifying a new transaction, so the actual args
    # connot be in the database (if they are, insert will fail).
    my ($xid,$ddid) = @{$dbh->{m_transformation}}{'xid','ddid'};
    return undef 
	unless $dbh->modify( q{insert into DERIVED(XID,PID,DDID,FLAG,POS) 
			       values (?,?,?,?,?)},
			     $xid, $pid, $ddid, $iox, $pos );
    print "-- info: inserted dep $xid => $pid @ $pos : $iox, $ddid\n";

    # return parameter id as success
    $pid;
}

sub internalStdio {
    # purpose: add stdio argument to PARAMETER
    #          also add dependency to current transaction
    # paramtr: $arg (IN): <lfn> to attach to a stdio stream when calling
    # returns: value of PID, undef in case of error
    my ( $dbh, $lfn, $std, $iox ) = @_;

    # sanity check    
    unless ( defined $dbh->{m_transformation} && defined $dbh->{m_transformation}->{xid} ) {
	print "-- error: no transaction to attach\n";
	return undef;
    }

    # there can be only one
    if ( defined $dbh->{m_transformation}->{$std} ) {
	print( "-- error: $std already defined as ", 
	       $dbh->{m_transformation}->{$std}, "\n" );
	return undef;
    } else {
	$dbh->{m_transformation}->{$std} = $lfn;
    }

    # return pid or undef on error
    print "-- debug: lfn=$lfn, iox=$iox\n";
    internalArg( $dbh, $lfn, $iox );
}

sub myStdIn {
    # purpose: add stdio argument to PARAMETER
    #          also add dependency to current transaction
    # paramtr: $arg (IN): <lfn> to attach to stdin when calling
    # returns: value of PID, undef in case of error
    internalStdio( @_[0..1], 'stdin', 'I' );
}

sub myStdOut {
    # purpose: add stdio argument to PARAMETER
    #          also add dependency to current transaction
    # paramtr: $arg (IN): <lfn> to attach to stdout when calling
    # returns: value of PID, undef in case of error
    internalStdio( @_[0..1], 'stdout', 'O' );
}

sub myStdErr {
    # purpose: add stdio argument to PARAMETER
    #          also add dependency to current transaction
    # paramtr: $arg (IN): <lfn> to attach to stderr when calling
    # returns: value of PID, undef in case of error
    internalStdio( @_[0..1], 'stderr', 'E' );
}

sub myArg {
    # purpose: add a non-file argument to STRING subclass of PARAMETER
    #          also add dependency to current transaction
    # paramtr: $arg (IN): <str> 
    # returns: value of PID, undef in case of error
    my $dbh = shift;
    my $arg = shift;

    # sanity check
    unless ( defined $dbh->{m_transformation} && defined $dbh->{m_transformation}->{xid} ) {
	print "-- error: no transaction to attach\n";
	return undef;
    }

    # return pid or undef on error
    print "-- debug: str=$arg, iox='-'\n";
    internalArg( $dbh, $arg, '-' );
}

sub myFile {
    # purpose: add a pfn to REPLICA_CATALOG and lfn to PARAMETER
    #          also adds dependency between xform and parameter
    # paramtr: $arg (IN): [i|o] <lfn>
    # returns: value of PID, undef in case of error
    my $dbh = shift;
    my $arg = shift;

    # sanity check    
    unless ( defined $dbh->{m_transformation} && defined $dbh->{m_transformation}->{xid} ) {
	print "-- error: no transaction to attach\n";
	return undef;
    }

    # extract flags and positional args
    my ($iox,$lfn) = split /\s+/,$arg,2;

    # return pid or undef on error
    print "-- debug: lfn=$lfn, iox=$iox\n";
    internalArg( $dbh, $lfn, $iox );
}

sub myRC {
    # purpose: add a mapping from a lfn to a pfn, may be used repeatedly
    # paramtr: $arg (IN): <lfn> <pfn> # PFN may contain spaces
    # returns: PID of LFN, undef in case of error
    # warning: outside block, function will commit changes
    my $dbh = shift;
    my ($lfn,$pfn,$pid) = split(/\s+/,shift,2);

    # phase 1: construct a new lfn PARAMETER, if it does not exist
    # maybe the file is already in the database?
    my $r = $dbh->fetch( q{select PID from PARAMETER where value = ?}, 
			   $lfn );

    if ( ref $r && @$r > 0 && ! $dbh->force ) {
	# logical file exists
	print "-- warn: multiple matching LFNs, taking first\n" 
	    if ( @$r > 1 );
	$pid = $r->[0]->[0];
	print "-- info: logical filename already known\n";
    } else {
	# data unknown, create new entry
	# obtain next sequence number to use as key
	$pid = $dbh->get_sequence('SEQ_PARAM') || return undef; # FIXME: die?
	print "-- info: using logical filename pid=$pid\n";

	# create transaction to insert logical filename
	return undef 
	    unless $dbh->modify( q{insert into PARAMETER(PID,VALUE) values (?,?)},
				 $pid, $lfn );
	print "-- info: inserted $pid => $lfn\n";
    }

    # phase 2: for given LFN PID, map PFN to it unless it exists
    # maybe the mapping is already in database?
    $r = $dbh->fetch( q{select * from REPLICA_CATALOG where PID=? and URI=?},
		      $pid, $pfn );
    if ( ref $r && @$r > 0 && ! $dbh->force ) {
	# lfn => pfn mapping exists
	print "-- warn: multiple matching RC entries, taking first\n" 
	    if ( @$r > 1 );
	print "-- info: mapping exists $lfn => $pfn, noop\n";
    } else {
	# data unknown, create new mapping entry
	return undef
	    unless $dbh->modify( q{insert into REPLICA_CATALOG(PID,URI) values (?,?)},
				 $pid, $pfn );
	print "-- info: inserted $lfn($pid) => $pfn\n";
    }

    # phase 3: outside block, commit changes
    unless ( defined $dbh->{m_transformation} && defined $dbh->{m_transformation}->{xid} ) {
	if ( my $result = $dbh->commit ) {
	    print( "-- info: committed $result modifications.\n" );
	} else {
	    print( "-- error: unable to commit $result modifications: ",
		   $dbh->handle->errstr, "\n" );
	}
    }
    
    # return PID on success
    $pid;
}

#
# --------------------------------------------------------------------
#

sub getAFilename ($$) {
    # purpose: map a lfn to a(ny) physical filename
    # paramtr: $dbh (IN): database object handle instance
    #          $lfn (IN): logical name of file to obtain
    # returns: undef in case of error, or scalar pfn containing data
    # warning: if there is no such mapping, function must fail 
    my ($dbh,$lfn) = @_;

    # find all associated physical filenames. If any exists, return it
    my $r = $dbh->fetch( q{select r.URI from REPLICA_CATALOG r, PARAMETER p
			   where r.PID=p.PID and p.VALUE=?}, 
			 $lfn );
    unless ( ref $r && @$r > 0 ) {
	print "-- error: there are no associated filenames!\n";
	return undef;
    }
    
    # check each file for existance (FIXME: readability for user)
    foreach my $fn ( @$r ) {
	print "-- debug: checking file ", $fn->[0], "\n"
	    if $dbh->debug->{get};
	print "-- info: ", $fn->[0], " matches $lfn\n";
	if ( -r $fn->[0] ) {
	    print "-- info: ", $fn->[0], " is readable to me\n";
	    return $fn->[0];
	}
    }

    # postcondition: There is at least one physical filename associated,
    # but it does not show up on the disks.
    #$r->[0]->[0];
    [ map { $_->[0] } @$r ];
}

# function prototype for recursion
sub myGet ($$);

# function implementation
sub myGet($$) {
    # purpose: recursively extract files from dependencies
    # paramtr: $dbh (IN): database object handle instance
    #          $lfn (IN): parameter id of file to obtain
    #          $cnt (IO): DAG node labeller
    # returns: undef in case of error, pfn containing data otherwise
    my ($dbh,$lfn) = @_;

    # step 1: find all associated physical filenames. If any exists, return it
    my $r = getAFilename($dbh,$lfn);
    return $r if ( defined $r and ! ref $r and ! $dbh->force );
    return undef unless defined $r;
    # postcondition: There is at least one physical filename associated,
    # but it does not show up on the disks. If there is a file on disk,
    # we would have return with above statement.
    # postcondition: In force mode, existing filenames will be ignored.
    print "-- warn: more than one pfn($lfn), using first\n"
	if ( ref $r && @$r > 1 );
    my $pfn = ( ref $r ? $r->[0] : $r );

    # step 2: obtain parameter id for logical filename
    $r = $dbh->fetch( q{select PID from PARAMETER where VALUE=?}, $lfn );
    unless ( ref $r && @$r > 0 ) { 
	print "-- error: dependency is unknown to me\n";
	return undef;
    }
    print "-- warn: multiple pid($lfn), using first one!\n" 
	if ( @$r > 1 );
    # postcondition: there is at least one PID for the LFN
    my $pid = $r->[0]->[0];

    # step 3: select all functions which contain this filename as output
    my $f = $dbh->fetch( q{select XID,DDID from DERIVED where PID=? and 
			   (FLAG='o' or FLAG='O' or FLAG='E')},
			 $pid );

    unless ( ref $f && @$f > 0 ) {
	print "-- error: there is no function to create parameter $pid\n";
	return undef;
    }
    print "-- warn: more than one production possible, using first\n"
	if @$f > 1;
    # postcondition: there is at least one function to create the file

    # obtain XID,DDID of function to use for creating the file.
    my ($xid,$ddid) = @{$f->[0]};
    print "-- debug: xid=$xid, ddid=$ddid\n";

    # step 4: collect information about the transformation
    my $x = $dbh->fetch( q{select XID,EXECUTABLE from TRANSFORMATION 
			   where XID=?}, $xid );
    unless ( ref $x && @$x > 0 ) {
	print "-- error: missing specification for function $xid\n";
	return undef;
    }
    # postcondition: XID is a key attribute, there can be only one spec
    print "-- debug: |", join('|',@{$x->[0]}), "|\n" 
	if $dbh->debug->{get};

    # step 5: collect all arguments for the actual transformation ddid
    my $in = $dbh->fetch( q{select d.PID,POS,FLAG,p.VALUE 
			    from DERIVED d, PARAMETER p 
			    where XID=? and DDID=? and d.PID=p.PID
			    order by POS}, $xid, $ddid );

    # functions may exist which are generating sources, ie no input files
    if ( ! ref $in || $DBI::err != 0 ) {
	print "-- error: database selection failed\n";
	return;
    }
    if ( $dbh->debug->{get} ) {
	foreach my $i ( @$in ) {
	    print '-- debug: |', join('|',@$i), "|\n";
	}
	print "-- debug: no input files\n" if @$in == 0;
    }

    # step 6: construct argument string for call
    my (@arguments,$postfix);
    foreach my $arg ( @$in ) {
	if ( $arg->[2] eq '-' ) {
	    # regular non-file parameter
	    push @arguments, $arg->[3];
	} elsif ( lc($arg->[2]) eq 'i' ) {
	    # input via file or stdin
	    # file existence check will be done first thing within internalGet
	    my $fnp = myGet( $dbh, $arg->[3] );
	    return undef unless defined $fnp;
	    if ( $arg->[2] eq 'i' ) {
		push @arguments, $fnp;
	    } else { 
		$postfix .= "<$fnp ";
	    }
	} elsif ( lc($arg->[2]) eq 'o' or $arg->[2] eq 'E' ) {
	    # output via file, stdout or stderr
	    # output file: get the appropriate physical filename
	    my $fnp;
	    if ( $lfn ne $arg->[3] ) {
		# new output file, other than the one we looked at previously
		my $r = getAFilename( $dbh, $arg->[3] );
		if ( defined $r ) {
		    $fnp = ref $r ? $r->[0] : $r;
		} else {
		    print "-- error: unable to map output file\n";
		    return undef;
		}
	    } else {
		# we know about this output file, because we asked for it 
		$fnp = $pfn;
	    }
	    if ( $arg->[2] eq 'o' ) {
		push @arguments, $fnp;
	    } else {
		$postfix .= '2' if $arg->[2] eq 'E';
		$postfix .= ">$fnp ";
	    }
	} else {
	    # this must not happen
	    print( '-- error: illegal combination of flag=',
		   $arg->[2], ' and pos=', $arg->[1], "\n" );
	    return undef;
	}
    }

    # step 7: execute remainder
    my $arguments = join ' ', @arguments;
    print "-- debug: ", $x->[0]->[1], ' ', $arguments, ' ', $postfix, "\n";
    unless ( system( $x->[0]->[1] . " $arguments $postfix" ) == 0 ) {
	# error during execution of system() occurred
	my $rv = $?;
	my $msg = sprintf( "function %ld failed: exit=%d, signal=%02x", 
			   $x->[0]->[0], $rv >> 8, $rv & 127 );
	$msg .= " (core dumped)" if ( $rv & 128 );
	print "-- error: ", $msg, "\n";
	return undef;
    }

    # result is the produced physical filename
    $pfn;
}


sub lfn2pfn ($$) {
    # purpose: map a lfn to a(ny) physical filename
    # paramtr: $dbh (IN): database object handle instance
    #          $lfn (IN): logical name of file to obtain
    # returns: undef in case of error, or vector of pfns
    # warning: if there is no such mapping, function must fail 
    my ($dbh,$lfn) = @_;

    # find all associated physical filenames. If any exists, return it
    my $r = $dbh->fetch( q{select r.URI from REPLICA_CATALOG r, PARAMETER p
			   where r.PID=p.PID and p.VALUE=?}, 
			 $lfn );
    unless ( ref $r && @$r > 0 ) {
	print "-- error: there are no associated filenames!\n";
	return undef;
    }

    # postcondition: There is at least one physical filename associated,
    [ map { $_->[0] } @$r ];
}

sub internalDag ($$$);
sub internalDag($$$) {
    # purpose: recursively extract DAG from dependencies
    # paramtr: $dbh (IN): database object handle instance
    #          $lfn (IN): parameter id of file to obtain
    #          $cnt (IO): DAG handler
    # returns: undef in case of error, ddid of current xform otherwise
    my ($dbh,$lfn,$dagger) = @_;

    # step 2: obtain parameter id for logical filename
    my $r = $dbh->fetch( q{select PID from PARAMETER where VALUE=?}, $lfn );
    unless ( ref $r && @$r > 0 ) { 
	print "-- error: dependency is unknown to me\n";
	return undef;
    }
    print "-- warn: multiple pid($lfn), using first one!\n" 
	if ( @$r > 1 );
    # postcondition: there is at least one PID for the LFN
    my $pid = $r->[0]->[0];

    # step 3: select all functions which contain this filename as output
    my $f = $dbh->fetch( q{select XID,DDID from DERIVED where PID=? and 
			   (FLAG='o' or FLAG='O' or FLAG='E')},
			 $pid );

    unless ( ref $f && @$f > 0 ) {
	print "-- error: there is no function to create parameter $pid\n";
	return undef;
    }
    print "-- warn: more than one production possible, using first\n"
	if @$f > 1;
    # postcondition: there is at least one function to create the file

    # obtain XID,DDID of function to use for creating the file.
    my ($xid,$ddid) = @{$f->[0]};
    print "-- debug: xid=$xid, ddid=$ddid\n";

    # step 4: collect information about the transformation
    my $x = $dbh->fetch( q{select XID,EXECUTABLE from TRANSFORMATION 
			   where XID=?}, $xid );
    unless ( ref $x && @$x > 0 ) {
	print "-- error: missing specification for function $xid\n";
	return undef;
    }
    # postcondition: XID is a key attribute, there can be only one spec
    print "-- debug: |", join('|',@{$x->[0]}), "|\n" 
	if $dbh->debug->{dag};

    # step 5: collect all arguments for the actual transformation ddid
    my $in = $dbh->fetch( q{select d.PID,POS,FLAG,p.VALUE 
			    from DERIVED d, PARAMETER p 
			    where DDID=? and d.PID=p.PID
			    order by POS}, $ddid );

    # functions may exist which are generating sources, ie no input files
    if ( ! ref $in || $DBI::err != 0 ) {
	print "-- error: database selection failed\n";
	return;
    }
    if ( $dbh->debug->{dag} ) {
	foreach my $i ( @$in ) {
	    print '-- debug: |', join('|',@$i), "|\n";
	}
	print "-- debug: no input files\n" if @$in == 0;
    }

    # step 6: construct argument string for abstract DAG
    my (@arguments,%postfix);
    foreach my $arg ( @$in ) {
	if ( $arg->[2] eq '-' ) {
	    # regular non-file parameter
	    push @arguments, $arg->[3];
	} elsif ( lc($arg->[2]) eq 'i' ) {
	    # input via file or stdin
	    # file existence check will be done first thing within internalGet
	    my $sub_ddid = internalDag( $dbh, $arg->[3], $dagger );
	    return undef unless defined $sub_ddid;
	    # add parent ddid to cache
	    push( @{$dagger->{m_parents}->{$ddid}}, $sub_ddid );
	    # set up lfn parameters
	    my $pfn = $dagger->check($arg->[3]) || 
		lfn2pfn($dbh,$arg->[3])->[0] || '@{' . $arg->[3] . '}';
	    $dagger->store($arg->[3],$pfn);
	    if ( $arg->[2] eq 'i' ) {
		push @arguments, $pfn;
	    } else { 
		$postfix{stdin} = $pfn;
	    }
	} elsif ( lc($arg->[2]) eq 'o' or $arg->[2] eq 'E' ) {
	    # output via file, stdout or stderr
	    # output file: get the appropriate physical filename
	    my $pfn = $dagger->check($arg->[3]) || 
		lfn2pfn($dbh,$arg->[3])->[0] || '@{' . $arg->[3] . '}';
	    $dagger->store($arg->[3],$pfn);
	    if ( $arg->[2] eq 'o' ) {
		push @arguments, $pfn;
	    } else {
		if ( $arg->[2] eq 'E' ) {
		    $postfix{stderr} = $pfn;
		} else {
		    $postfix{stdout} = $pfn;
		}
	    }
	} else {
	    # this must not happen
	    print( '-- error: illegal combination of flag=',
		   $arg->[2], ' and pos=', $arg->[1], "\n" );
	    return undef;
	}
    }

    # step 7: generate DAG node job description
    my $nodename = $dagger->name($ddid);
    unless ( $dagger->{m_visited}{$nodename} ) {
	$dagger->{m_visited}{$nodename} = "$nodename.sub";

	# generate DAG node submission configuration
	my $node = new IO::File "$nodename.sub", O_WRONLY | O_CREAT;
	if ( defined $node ) {
	    # generate data for configuration
	    print $node "#\n# Filename: $nodename.sub\n#\n";
	    print $node "Universe     = vanilla\n";
	    print $node 'Executable   = ', $x->[0]->[1], "\n";
	    print $node 'Log          = ', $dagger->log, "\n";
	    print $node 'Input        = ', $postfix{stdin}, "\n"  if $postfix{stdin};
	    print $node 'Output       = ', $postfix{stdout}, "\n" if $postfix{stdout};
	    print $node 'Error        = ', $postfix{stderr}, "\n" if $postfix{stderr};
	    print $node 'Arguments    = ', join(' ',@arguments), "\n" if @arguments > 0;
	    print $node "Notification = NEVER\n";
	    print $node "Queue\n";
	
	    # done
	    undef $node;
	} else {
	    print "-- error: unable to generate $nodename.sub: $!\n";
	}
    }

    # result is the produced used derivation index
    $ddid;
}

sub myDag ($$) {
    # purpose: recursively extract files from dependencies
    # paramtr: $dbh (IN): database object handle instance
    #          $lfn (IN): parameter id of file to obtain
    # returns: undef in case of error, pfn containing data otherwise
    my ($dbh,$arg) = @_;
    my ($dag,$lfn) = split /\s+/,$arg,2;

    my $fh = new IO::File "$dag.dag", O_WRONLY | O_CREAT;
    if ( defined $fh ) {
	# create new object
	my $dagger = Dagger->new( "$dag.log" );

	# create DAG and let it create job descriptions
	my $r = internalDag( $dbh, $lfn, $dagger );

	# dump job descriptions
	foreach my $ddid ( sort keys %{$dagger->{m_visited}} ) {
	    $fh->print( "Job $ddid ", $dagger->{m_visited}->{$ddid}, "\n" );
	}

	# dump dependencies
	foreach my $ddid ( sort keys %{$dagger->{m_parents}} ) {
	    $fh->print( 'PARENT ', join( ' ', map { $dagger->name($_) } 
					 @{$dagger->{m_parents}->{$ddid}} ),
			' CHILD ', $dagger->name($ddid), "\n" );
	}

	undef $fh;		# close()
	return $r;
    } else {
	print "-- error: unable to open $dag.dag: $!\n";
	return undef;
    }
}

#
# ---------------------------------------------------------------
#
sub mySet {
    # purpose: set a local configuration value
    my $dbh = shift;
    my ($key,$value) = split(/\s+/,shift(),2);

    # $key must follow identifier semantics
    unless ( $key =~ /^[A-Za-z_.-][A-Za-z0-9_.-]*/ ) {
	# complain
	print "-- error: $key is not a valid identifier\n";
	return undef;
    }

    # set or overwrite a value
    $dbh->{m_config}->{$key} = $value;

    # return result
    1;
}

sub myUnset {
    # purpose: reset a local configuration value
    my ($dbh,$key) = @_;

    # $key must follow identifier semantics
    unless ( $key =~ /^[A-Za-z_.-][A-Za-z0-9_.-]*/ ) {
	# complain
	print "-- error: $key is not a valid identifier\n";
	return undef;
    }

    # remove a value
    if ( exists $dbh->{m_config}->{$key} ) {
	 delete $dbh->{m_config}->{$key};
	 return 1;
     }

    # return result
    0;
}

sub myShow {
    # purpose: show local configuraiton value(s)
    my $dbh = shift;
    my @key = ( $_[0] =~ /^\s*$/ ? 
		keys %{$dbh->{m_config}} : 
		split(/\s+/,shift()) );
    foreach my $key ( @key ) {
	if ( exists $dbh->{m_config}->{$key} ) {
	    print( 'set ', $key, ' ', $dbh->{m_config}->{$key}, "\n" );
	} else {
	    print "-- info: $key is not set\n";
	}
    }

    # result
    0+@key;
}

sub myLoad {
    # purpose: read settings from a file, also works as "include" option
    # returns: number of valid lines that matter.
    my ($dbh,$fn) = @_;
    processFile($dbh,$fn);
}

sub mySave {
    my ($dbh,$fn) = @_;
    
    my $fh = new IO::File $fn, O_WRONLY | O_CREAT;
    if ( defined $fh ) {
	foreach my $key ( sort keys %{$dbh->{m_config}} ) {
	    print $fh 'set ', $key, ' ', $dbh->{m_config}->{$key}, "\n";
	}
	undef $fh;
    } else {
	print "-- error: unable to open $fh: $!\n";
	return undef;
    }
    $fn;
}
